/* ScummVM - Scumm Interpreter
 * Copyright (C) 2001/2002 The ScummVM project
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Header$
 *
 */

#ifndef __BRENDERER_H_
#define __BRENDERER_H_

#include "config.h"

#include "renderer.h"
#include "palette.h"
#include "rect.h"

/*! 	@brief base class for easily creating ::renderer instances

	This class implements some function available in the ::renderer abstract class, so that 
	creation of subclasses of ::renderer is easier.
*/
class base_renderer : public renderer {
private:
	palette _pal;		//!< The current palette
	char * _data;		//!< The current frame buffer
	int _frame;			//!< The current frame number
	int _nbframes;		//!< The number of frames in the animation
	int _width;			//!< The current frame's width
	int _height;		//!< The current frame's height
	const char * _fname;	//!< The filename of the animation being played
protected:
	virtual void save(int frame = -1) = 0;

protected:
	const char * get_filename() const { return _fname; };	//!< accessor for animation filename
	int get_nbframes() const { return _nbframes; };	//!< accessor for number of frames
	int get_width() const { return _width; };	//!< accessor for current width
	int get_height() const { return _height; };	//!< accessor for current height
	const palette & pal() const { return _pal; };	//!< accessor for current palette
	const char * data() const { return _data; };	//!< accessor for current frame buffer
	void clean();	//!< memory cleanup (deletes frame buffer)
	void set_frame(int f) { _frame = f; };	//!< allows to change the frame number
public:
	int get_frame() const { return _frame; };	//!< accessor for current frame number
	base_renderer();
	virtual ~base_renderer();
	
	virtual bool init_frame(const point & size);
	virtual char * lock_frame(int frame);
	virtual bool unlock_frame();
	virtual bool flip_frame();
	virtual bool set_palette(const palette & pal);
	virtual bool start_decode(const char * fname, int version, int nbframes) { _fname = fname; _nbframes = nbframes; return true; }
	virtual mixer * get_mixer() { return 0; };
	virtual bool premature_close() { return false; };
};

/*! 	@brief A null ::renderer

	This class completely implements ::renderer, without actually doing anything.
	This class is useful for performance measurements.
*/
class null_renderer : public base_renderer {
protected:
	void save(int frame = -1) {};
public:
	null_renderer() {};
	virtual ~null_renderer() {};
	bool wait(int ms) { return true; };
};

#endif
