/* ScummVM - Scumm Interpreter
 * Copyright (C) 2001/2002 The ScummVM project
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Header$
 *
 */

#ifndef __CHANNEL_H_
#define __CHANNEL_H_

#include "config.h"

#ifdef DEBUG
# ifndef NO_DEBUG_CHANNEL
#  define DEBUG_CHANNEL
# endif
#else
# ifdef DEBUG_CHANNEL
#  error DEBUG_CHANNEL defined without DEBUG
# endif
#endif

class chunck;
class cont_chunck;
	
/*! 	@brief interface for a sound channel (a track)

	This is the interface for sound channels. 
*/
class channel {
public:
	virtual ~channel() {};
	// called by the smush_player
	virtual bool append_data(chunck & b, int size) = 0;
	virtual bool set_parameters(int, int, int, int) = 0;
	virtual bool check_parameters(int, int, int, int, int) = 0;
	// called by the mixer
	virtual bool is_terminated() const = 0;
	virtual int available_sound_data() const = 0;
	virtual void get_sound_data(short * sound_buffer, int size) = 0; // size is in sample 
	virtual void get_sound_data(char * sound_buffer, int size) = 0;
	virtual bool get_parameters(int &rate, bool &stereo, bool &is_16bit) = 0;
	virtual int get_track_identifier() const = 0;
	
};

class saud_channel : public channel {
private:
	int _track;				//!< The track identifier
	int _nbframes;			//!< number of frames of the track (unused)
	int _data_size;			//!< the size of the sound buffer
	int _frequency;			//!< the frequency target of the track (always 22050)
	bool _in_data;			//!< are we processing data ?
	bool _mark_reached;		//!< set to \c true when the SMRK tag is reached
	int _flags;				//!< current flags of the track (unused)
	int _volume;			//!< the current track volume
	int _balance;			//!< the current track balance
	int _index;				//!< the current PSAD index (for coherency checking)
	short _voltable[2][256];	//!< the precalculated volume table (stereo 16 bits)
	unsigned char * _tbuffer;	//!< data temporary buffer
	int _tbuffer_size;			//!< temporary buffer size
	unsigned char * _sbuffer;	//!< sound buffer
	int _sbuffer_size;			//!< sound buffer size
protected:
	void handleStrk(chunck & c);
	void handleSmrk(chunck & c);
	void handleShdr(chunck & c);
	bool handleSubTags(int & offset);
	bool processBuffer();
	void recalcVolumeTable();
public:
	saud_channel(int track, int freq);
	virtual ~saud_channel();
	bool is_terminated() const;
	bool set_parameters(int duration, int flags, int vol1, int vol2);
	bool check_parameters(int index, int duration, int flags, int vol1, int vol2);
	bool append_data(chunck & b, int size);
	int available_sound_data() const;
	void get_sound_data(short * sound_buffer, int size);
	void get_sound_data(char * sound_buffer, int size) { error("16bit request for SAUD channel should never happen"); };
	bool get_parameters(int &rate, bool &stereo, bool &is_16bit) { 
		rate = _frequency;
		stereo = true;
		is_16bit = true;
		return true;
	};
	virtual int get_track_identifier() const { return _track; };
};

/*! 	@brief class for a IACT sound ::channel (a The Dig track)

	This class implements a channel specifically for The Dig.

	\bug for unknown reason, some sound have a too long duration, or repeat themselves.
*/
class imuse_channel : public channel {
private:
	int _track;				//!< the track number
	unsigned char * _tbuffer;	//!< data temporary buffer
	int _tbuffer_size;			//!< temporary buffer size
	unsigned char * _sbuffer;	//!< sound buffer
	int _sbuffer_size;			//!< sound buffer size
	int _srbuffer_size;
	int _frequency;			//!< the target frequency of the ::mixer
	int _data_size;			//!< remaining size of sound data in the iMUS buffer
	bool _in_data;


	int _bitsize;			//!< the bitsize of the original data
	int _rate;				//!< the sampling rate of the original data
	int _channels;			//!< the number of channels of the original data

	// The following are parameters only kept to be consistent with what the file provides... They are not used.
	//~ int _size;
	//~ int _nbframes;
	//~ int _unk1;
	//~ int _unk2;
	//~ int _index;
	//~ int _imuse_start;
	//~ int _imuse_length;
	//~ int _imuse_stop;
protected:
	void decode();
	bool processBuffer();
	bool handleMap(chunck &);
	bool handleFormat(chunck &);
	bool handleText(chunck &);
	bool handleRegion(chunck &);
	bool handleStop(chunck &);
	bool handleSubTags(int & offset);
public:
	imuse_channel(int track, int freq);
	virtual ~imuse_channel();
	bool is_terminated() const;
	bool set_parameters(int nbframes, int size, int unk1, int unk2);
	bool check_parameters(int index, int nbframes, int size, int unk1, int unk2);
	bool append_data(chunck & b, int size);
	int available_sound_data() const;
	void get_sound_data(short * sound_buffer, int size);
	void get_sound_data(char * sound_buffer, int size);
	bool get_parameters(int &rate, bool &stereo, bool &is_16bit) {
		rate = _frequency;
		stereo = (_channels == 2);
		is_16bit = (_bitsize > 8);
		return true;
	};
	virtual int get_track_identifier() const { return _track; };
};

#endif
