/* ScummVM - Scumm Interpreter
 * Copyright (C) 2001/2002 The ScummVM project
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Header$
 *
 */

#include <stdafx.h>
#include "chunck.h"

#include <stdio.h> // for FILE, fopen, fclose, fseek and ftell
#include <string.h> // for memcpy

/*!	@brief very small and fast wrapper for a ifstream.

	implements reference counting, so that ::file_chunck does not leak memory !
*/
class file_ptr {
	char * _filename;
	FILE * _ifs;
	int _refcount;
	int _cur_pos;
public:
	file_ptr(const char * fname) : _refcount(1), _cur_pos(0) {
		debug(9, "file_ptr created for %s", fname);
		_filename = strdup(fname);
		_ifs  = fopen(fname, "rb");
		if(_ifs == NULL) error("file_ptr unable to read file \"%s\"", fname);
	}
	~file_ptr() {
		debug(9, "file_ptr destroyed for %s", _filename);
		free(_filename);
		fclose(_ifs);
	}
	int tell() {
		return _cur_pos;
	}
	bool seek(int pos) {
		if(pos != _cur_pos) {
			fseek(_ifs, pos, SEEK_SET);
			_cur_pos = pos;
		}
		return true;
	}
	bool read(void * ptr, int size) {
		fread(ptr, size, 1, _ifs);
		_cur_pos += size;
		return true;
	}
	void inc_ref() {
		_refcount++;
	}
	void dec_ref() {
		if(--_refcount == 0)
			delete this;
	}
};

const char * chunck::chunck_string(chunck::type t) {
	static char data[5];
	data[0] = (char)((t >> 24) & 0xFF);
	data[1] = (char)((t >> 16) & 0xFF);
	data[2] = (char)((t >> 8) & 0xFF);
	data[3] = (char)((t >> 0) & 0xFF);
	data[4] = 0;
	return data;
}


file_chunck::file_chunck() : _data(0), _type(0), _size(0), _cur_pos(0) {
}

file_chunck::~file_chunck() {
	if(_data) _data->dec_ref();
}

file_chunck::file_chunck(const char * fname) {
	_data = new file_ptr(fname);
	_data->read(&_type, 4);
	_type = TO_BE_32(_type);
	_data->read(&_size, 4);
	_size = TO_BE_32(_size);
	_offset = _data->tell();
	_cur_pos = 0;
}

chunck::type file_chunck::get_type() const { 
	return _type; 
}

unsigned int file_chunck::get_size() const { 
	return _size; 
}

chunck * file_chunck::sub_block() {
	file_chunck * ptr = new file_chunck;
	ptr->_data = _data;
	_data->inc_ref();
	_data->seek(_offset+_cur_pos);
	unsigned int temp;
	_data->read(&temp, 4);
	ptr->_type = TO_BE_32(temp);
	_data->read(&temp, 4);
	ptr->_size = TO_BE_32(temp);
	ptr->_offset = _offset + _cur_pos + 8;
	ptr->_cur_pos = 0;
	seek(8 + ptr->get_size());
	return ptr;
}

bool file_chunck::eof() const { 
	return _cur_pos >= _size; 
}

unsigned int file_chunck::tell() const { 
	return _cur_pos; 
}

bool file_chunck::seek(int delta, seek_type dir) {
	switch(dir) {
		case seek_cur:
			_cur_pos += delta;
			break;
		case seek_start:
			if(delta < 0) error("invalid seek request");
			_cur_pos = (unsigned int)delta;
			break;
		case seek_end:
			if(delta > 0 || (_size + delta) < 0) error("invalid seek request");
			_cur_pos = (unsigned int)(_size + delta);
			break;
	}
	if(_cur_pos > _size) {
		error("invalid seek request : %d > %d (delta == %d)", _cur_pos, _size, delta);
	}
	return true;
}

bool file_chunck::read(void * buffer, unsigned int size) {
	if(size <= 0 || (_cur_pos + size) > _size) error("invalid buffer read request");
	_data->seek(_offset + _cur_pos);
	_data->read(buffer, size);
	_cur_pos += size;
	return true;
}

char file_chunck::get_char() {
	if(_cur_pos >= _size) error("invalid char read request");
	_data->seek(_offset + _cur_pos);
	char buffer;
	_data->read(&buffer, sizeof(buffer));
	_cur_pos+= sizeof(buffer);
	return buffer;
}

unsigned char file_chunck::get_byte() {
	if(_cur_pos >= _size) error("invalid byte read request");
	_data->seek(_offset + _cur_pos);
	unsigned char buffer;
	_data->read(&buffer, sizeof(buffer));
	_cur_pos+= sizeof(buffer);
	return buffer;
}

short file_chunck::get_short() {
	//~ if(_cur_pos >= _size-1) error("invalid short read request");
	unsigned short buffer = get_word();
	return *((short*)&buffer);
}

unsigned short file_chunck::get_word() {
	if(_cur_pos >= _size-1) error("invalid word read request");
	_data->seek(_offset + _cur_pos);
	unsigned short buffer;
	_data->read(&buffer, sizeof(buffer));
	_cur_pos+= sizeof(buffer);
	return TO_LE_16(buffer);
}

unsigned int file_chunck::get_dword() {
	if(_cur_pos >= _size-3) error("invalid dword read request");
	_data->seek(_offset + _cur_pos);
	unsigned int buffer;
	_data->read(&buffer, sizeof(buffer));
	_cur_pos+= sizeof(buffer);
	return TO_LE_32(buffer);
}

cont_chunck::cont_chunck(char * data) {
	if(data == 0) error("chunck() called with NULL pointer");
	_type = (chunck::type)READ_BE_UINT32(data);
	_size = READ_BE_UINT32(data+4);
	_data = data + sizeof(chunck::type) + sizeof(unsigned int);
	_cur_pos = 0;
}

chunck::type cont_chunck::get_type() const { 
	return _type; 
}

unsigned int cont_chunck::get_size() const { 
	return _size; 
}

chunck * cont_chunck::sub_block() {
	cont_chunck * ptr = new cont_chunck(_data + _cur_pos);
	seek(sizeof(chunck::type) + sizeof(unsigned int) + ptr->get_size());
	return ptr;
}

bool cont_chunck::eof() const { 
	return _cur_pos >= _size; 
}

unsigned int cont_chunck::tell() const { 
	return _cur_pos; 
}

bool cont_chunck::seek(int delta, seek_type dir) {
	switch(dir) {
		case seek_cur:
			_cur_pos += delta;
			break;
		case seek_start:
			if(delta < 0) error("invalid seek request");
			_cur_pos = (unsigned int)delta;
			break;
		case seek_end:
			if(delta > 0 || (_size + delta) < 0) error("invalid seek request");
			_cur_pos = (unsigned int)(_size + delta);
			break;
	}
	if(_cur_pos > _size) {
		error("invalid seek request : %d > %d (delta == %d)", _cur_pos, _size, delta);
	}
	return true;
}

bool cont_chunck::read(void * buffer, unsigned int size) {
	if(size <= 0 || (_cur_pos + size) > _size) error("invalid buffer read request");
	memcpy(buffer, _data + _cur_pos, size);
	_cur_pos += size;
	return true;
}

char cont_chunck::get_char() {
	if(_cur_pos >= _size) error("invalid char read request");
	return _data[_cur_pos++];
}

unsigned char cont_chunck::get_byte() {
	if(_cur_pos >= _size) error("invalid byte read request");
	unsigned char * ptr = (unsigned char *)(_data + _cur_pos);
	_cur_pos += 1;
	return *ptr;
}

short cont_chunck::get_short() {
	if(_cur_pos >= _size-1) error("invalid short read request");
	unsigned short buffer = get_word();
	return *((short*)&buffer);
}

unsigned short cont_chunck::get_word() {
	if(_cur_pos >= _size-1) error("invalid word read request");
	unsigned short * ptr = (unsigned short *)(_data + _cur_pos);
	_cur_pos += 2;
	return READ_LE_UINT16(ptr);
}

unsigned int cont_chunck::get_dword() {
	if(_cur_pos >= _size-3) error("invalid dword read request");
	unsigned int * ptr = (unsigned int *)(_data + _cur_pos);
	_cur_pos += 4;
	return READ_LE_UINT32(ptr);
}
