/* ScummVM - Scumm Interpreter
 * Copyright (C) 2001/2002 The ScummVM project
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Header$
 *
 */

#include <stdafx.h>
#include "player.h"

#include "renderer.h"
#include "channel.h"
#include "chunck_type.h"
#include "rect.h"
#include "blitter.h"

#include <assert.h>
#include <stdlib.h> // for atoi
#include <stdio.h> // for FILE, fopen, fclose, fread, fseek, ftell
#include <string.h> // for strchr, strrchr
#include <ctype.h> // for isdigit

#ifndef max
#define max(x, y) ((x) > (y) ? (x) : (y))
#endif

const int WAIT = 100;

/*!	@brief parser and map of string resources

	This class implements a parser for the string resource format of SMUSH animations.
	It then allows the player to get the string corresponding to a particular identifier.
	
	@bug some of The Dig strings are not completely parsed (in titles)
*/
class string_resource {
private:
	static const int MAX_STRINGS = 200;
	struct {
		int id;
		char * string;
	} _strings[MAX_STRINGS];
	int _nb_strings;
	int _last_id;
	char * _last_string;
public:
	string_resource() : _nb_strings(0), _last_id(-1) {};
	~string_resource() {
		for(int i = 0; i < _nb_strings; i++) {
			delete []_strings[i].string;
		}
	}
	/*!	@brief parse the given buffer

		@param buffer the buffer that contain the resource (in lucasart format)
		@param length the length of the buffer

		@return \c true if the parsing went fine, \c false otherwise
	*/
	bool init(char * buffer, int length) {
		debug(9, "parsing string resources...");
		char * def_start = strchr(buffer, '#');
		while(def_start != NULL) {
			char * def_end = strchr(def_start, '\n');
			assert(def_end != NULL); // def_end is just before the start of the string [def_start,def_end] correspond to the definition text
			char * id_end = def_end;
			while(id_end >= def_start && !isdigit(*(id_end-1))) id_end--;
			assert(id_end > def_start);
			char * id_start = id_end;
			while(isdigit(*(id_start-1))) id_start--;
			// [id_start-id_end] is the id number
			char idstring[32];
			memcpy(idstring, id_start, id_end - id_start);
			idstring[id_end - id_start] = 0;
			int id = atoi(idstring);
			//~ assert(id != LONG_MIN && id != 0 && id != LONG_MAX);
			char * data_start = def_end;
			while(*data_start == '\n' || *data_start == '\r') data_start++;
			char * data_end = data_start;
			while(1) {
				if(data_end[-2] == '\r' && data_end[1] == '\n' && data_end[-1] == '\n' && data_end[0] == '\r')
					break;
				data_end++;
				if(data_end >= buffer+length) {
					data_end = buffer+length;
					break;
				}
			}
			data_end -= 2;
			assert(data_end > data_start);
			char * value = new char[data_end - data_start + 1];
			assert(value);
			memcpy(value, data_start, data_end - data_start);
			value[data_end - data_start] = 0;
#ifdef DEBUG
			debug(9, "Inserting (%s)%d == \"%s\"", idstring, id, value);
#endif				
			_strings[_nb_strings].id = id;
			_strings[_nb_strings].string = value;
			_nb_strings ++;
			def_start = strchr(data_end+2, '#');
		}
		return true;
	}
	/*!	@brief extract a string

		@param id the resource identifier

		@return the corresponding string.
	*/
	const char * get(int id) {
		if(id == _last_id) return _last_string;
		for(int i = 0; i < _nb_strings; i++)
		{
			if(_strings[i].id == id) {
				_last_id = id;
				_last_string = _strings[i].string;
				return _strings[i].string;
			}
		}
		warning("invalid string id : %d", id);
		_last_id = -1;
		_last_string = "unknown string";
		return _last_string;
	}
};

void smush_player::show(const char * p) {
	if(strcmp(p, "subtitles") == 0)
		_subtitles = true;
	else if(strcmp(p, "bgmusic") == 0)
		_bgmusic = true;
	else if(strcmp(p, "voices") == 0)
		_voices = true;
	else {
		int id = atoi(p);
		if(id < 0 || id > 36) error("invalid parameter to show");
		_skips[id] = true;
	}
}

void smush_player::hide(const char * p) {
	if(strcmp(p, "subtitles") == 0)
		_subtitles = false;
	else if(strcmp(p, "bgmusic") == 0)
		_bgmusic = false;
	else if(strcmp(p, "voices") == 0)
		_voices = false;
	else {
		int id = atoi(p);
		if(id < 0 || id > 36) error("invalid parameter to hide");
		_skips[id] = false;
	}
}


smush_player::smush_player(renderer * renderer, bool wait, bool sound) : 
							_version(-1), 
							_secondary_version(0),
							_sound_frequency(0),
							_nbframes(0), 
							_mixer(0),
							_renderer(renderer), 
							_strings(0),
							_frame_size(-1, -1), 
							_frame(0),
							_output_sound(sound),
							_wait(wait),
							_already_init(false),
							_codec37_called(false),
							_skipnext(false),
							_subtitles(true),
							_bgmusic(true),
							_voices(true)
{
	_fr[0] = _fr[1] = _fr[2] = _fr[3] = 0;
	assert(_renderer != 0);
}

smush_player::~smush_player() {
	clean();
	//~ if(_mixer) delete _mixer;
}

void smush_player::update_palette(void) { 
	_renderer->set_palette(_pal); 
}

void smush_player::clean() {
	if(_strings) delete _strings;
	if(_fr[0]) delete _fr[0];
	if(_fr[1]) delete _fr[1];
	if(_fr[2]) delete _fr[2];
	if(_fr[3]) delete _fr[3];
}

void smush_player::checkBlock(const chunck & b, chunck::type type_expected, unsigned int min_size) {
	if(type_expected != b.get_type()) {
		error("chunck type is different from expected : %d != %d", b.get_type(), type_expected);
	}
	if(min_size > b.get_size()) {
		error( "chunck size is inferior than minimum required size : %d < %d", b.get_size(), min_size);
	}
}

void smush_player::handleSoundBuffer(int track_id, int index, int max_frames, int flags, int vol, int bal, chunck & b, int size) {
	debug(6, "smush_player::handleSoundBuffer(%d)", track_id);
	if(!_voices && (flags & 128) == 128) return;
	if(!_bgmusic && (flags & 64) == 64) return;
	channel * c = _mixer->find_channel(track_id);
	if(c == 0) {
		c = new saud_channel(track_id,  _sound_frequency);
		_mixer->add_channel(c);
	}
	if(index == 0)
		c->set_parameters(max_frames, flags, vol, bal);
	else
		c->check_parameters(index, max_frames, flags, vol, bal);
	c->append_data(b, size);
}

void smush_player::handleSoundFrame(chunck & b) { 
	checkBlock(b, TYPE_PSAD);
	debug(6, "smush_player::handleSoundFrame()");
	if(!_output_sound) return;
	int track_id = b.get_word();
	int index = b.get_word();
	int max_frames = b.get_word();
	int flags = b.get_word();
	int vol = b.get_byte();
	int bal = b.get_char();
#ifdef DEBUG
	if(index == 0) {
		debug(5, "track_id == %d, max_frames == %d, %d, %d, %d", track_id, max_frames, flags, vol, bal);
	}
#endif
	int size = b.get_size() - 10;
	handleSoundBuffer(track_id, index, max_frames, flags, vol, bal, b, size);
}

void smush_player::handleSkip(chunck & b) { 
	checkBlock(b, TYPE_SKIP, 4);
	int code = b.get_dword();
	debug(6, "smush_player::handleSkip(%d)", code);
	if(code >= 0 && code < 37)
		_skipnext =_skips[code];
	else
		_skipnext =true;
}

void smush_player::handleStore(chunck & b) { 
	checkBlock(b, TYPE_STOR, 4);
	debug(6, "smush_player::handleStore()");
}

void smush_player::handleFetch(chunck & b) { 
	checkBlock(b, TYPE_FTCH, 6); 
	debug(6, "smush_player::handleFetch()");
}

void smush_player::handleImuseBuffer(int track_id, int index, int nbframes, int size, int unk1, int unk2, chunck & b, int bsize) { 
	channel * c = _mixer->find_channel(track_id);
	if(c == 0) {
		c = new imuse_channel(track_id, _sound_frequency);
		_mixer->add_channel(c);
	}
	if(index == 0)
		c->set_parameters(nbframes, size, unk1, unk2);
	else
		c->check_parameters(index, nbframes, size, unk1, unk2);
	c->append_data(b, bsize);
}

void smush_player::handleImuseAction8(chunck & b, int flags, int unknown, int track_id) {
	assert(flags == 46 && unknown == 0);
	int unknown2 = b.get_word();
	track_id |= unknown2 << 16;
	int index = b.get_word();
	int nbframes = b.get_word();
	int size = b.get_dword();
	int bsize = b.get_size() - 18;
	handleImuseBuffer(track_id, index, nbframes, size, unknown, unknown2, b, bsize);
}

void smush_player::handleImuseAction(chunck & b) { 
	checkBlock(b, TYPE_IACT, 8);
	debug(6, "smush_player::handleImuseAction()");
	if(!_output_sound) return;
	int code = b.get_word();
	int flags = b.get_word();
	int unknown = b.get_short();
	int track_id = b.get_word();
#ifdef DEBUG
	debug(5, "handleImuseAction(%d, %d, %d, %d)", code, flags, unknown, track_id);
#endif
	switch(code) {
		case 8:
			handleImuseAction8(b, flags, unknown, track_id);
			break;
#ifdef DEBUG
		default:
			{
				debug(9, "%5.5d %d %8.8d %4.4d", track_id, flags, unknown);
				// because debug appends a carriage return, I can't do this like I used to...
				//~ int i = 1;
				//~ while(!b.eof()) {
					//~ int value = b.get_byte();
					//~ debug("%3.3d ", value);
					//~ if(i++ == 80) { i = 0; debug("\n"); }
				//~ }
				//~ if(i) debug("\n");
			}
#endif
	}
}

//~ #include <iostream>

void smush_player::handleTextResource(chunck & b) { 
	checkBlock(b, TYPE_TRES, 18); 
	int pos_x = b.get_short();
	int pos_y = b.get_short();
	int flags = b.get_short();
	int left = b.get_short();
	int top = b.get_short();
	int width = b.get_short();
	int height = b.get_short();
	int unk2 = b.get_word();
	int string_id = b.get_word();
	debug(6, "smush_player::handleTextResource(%d)", string_id);
	if(!_strings) return;
	// if subtitles disabled and bit 3 is set, then do not draw
	if((!_subtitles) && ((flags & 8) == 8)) return;
	const char * str = _strings->get(string_id);

	font_renderer * fr = _fr[0];
	int color = 15;
	while(*str == '/') str++; // For Full Throttle text resources
	while(str[0] == '^') {
		switch(str[1]) {
		case 'f':
			{
#if 0
				// This cause trouble if the next character is a digit.
				int id = atoi(str+2);
#else
				// assume ASCII like character set...
				int id = str[3] - '0';
#endif
				str += 4;
				fr = _fr[id]; 
			} break;
		case 'c':
			{
				//~ int id = atoi(str+2);
				color = str[4] - '0' + 10 *(str[3] - '0');
				str += 5;
			} break;
		default:
			error("invalid escape code in text string");
		}
	}
	assert(fr != 0);
	fr->set_color(color);
	if(!_cur_buffer) { _cur_buffer = _renderer->lock_frame(_frame); }
	if(flags == 0 || flags == 4) {
		fr->draw_string_absolute(str, _cur_buffer, _frame_size, pos_x, pos_y);
	} else {
		fr->draw_string_centered(str, _cur_buffer, _frame_size, max(pos_y, top), left, width, pos_x);
	}
}

void smush_player::readPalette(palette & out, chunck & in) {
	unsigned char buffer[768];
	in.read(buffer, 768);
	out = palette(buffer);
}

void smush_player::handleDeltaPalette(chunck & b) { 
	checkBlock(b, TYPE_XPAL);
	debug(6, "smush_player::handleDeltaPalette()");
	if(b.get_size() == 768*3+4) {
		int unk1, num;
		unk1 = b.get_word();
		num = b.get_word();
		for(int i = 0; i < 768; i++)
		{
			_delta_pal[i] = b.get_word();
		}
		readPalette(_pal, b);
		update_palette();
	} else if(b.get_size() == 6) {
		int unk1, num, unk2;
		unk1 = b.get_word();
		num = b.get_word();
		unk2 = b.get_word();
		for(int i = 0; i < 256; i++)
		{
			_pal[i].delta(_delta_pal+3*i);
		}
		update_palette();
	} else {
		error("wrong size for DeltaPalette");
	}
}

void smush_player::handleNewPalette(chunck & b) { 
	checkBlock(b, TYPE_NPAL, 768);
	debug(6, "smush_player::handleNewPalette()");
	readPalette(_pal, b);
	update_palette();
}

void smush_player::decodeCodec(chunck & b, const rect & r, decoder & codec) {
	assert(_cur_buffer);
	blitter blit(_cur_buffer, _frame_size, r);
	codec.decode(blit, b);
}

void smush_player::init_size(const rect & r, bool always, bool transparent) {
	if(_codec37_called) _already_init = true;
		
	if(!_already_init || _frame_size.get_x() < r.right() || _frame_size.get_y() < r.bottom() || always) {
		if(_cur_buffer) { _renderer->unlock_frame(); _cur_buffer = 0; }
		_frame_size = r.bottomright();
		_renderer->init_frame(_frame_size);
	}
	if(_cur_buffer) { _renderer->unlock_frame(); _cur_buffer = 0; }
	_cur_buffer = _renderer->lock_frame(_frame);
	if(!_already_init && transparent) {
		memset(_cur_buffer, 0, _frame_size.get_x()*_frame_size.get_y());
	}
	_codec1.init_size(_frame_size, r);
	_codec37.init_size(_frame_size, r);
	_codec44.init_size(_frame_size, r);
	_codecd.init_size(_frame_size, r);
	_already_init = true;
}

void smush_player::handleFrameObject(chunck & b) { 
	checkBlock(b, TYPE_FOBJ, 14);
	if(_skipnext) {
		_skipnext = false;
		return;
	}
	int codec = b.get_word();
	debug(6, "smush_player::handleFrameObject(%d)", codec);
	unsigned short left = b.get_word();
	unsigned short top = b.get_word();
	unsigned short width = b.get_word();
	unsigned short height = b.get_word();
	rect r(left, top, left+width, top+height);
	unsigned short data[2];
	data[1] = b.get_word();
	data[0] = b.get_word();
#ifdef DEBUG
	debug(5, "Frame pos : %d, %d", left, top);
	debug(5, "Frame size : %dx%d", width, height);
	debug(5, "Codec : %d", codec);
#endif
	switch (codec) {
	case 3:
	case 1:
		init_size(r, false, true);
		decodeCodec(b, r, _codec1);
		break;
	case 37:
		assert(left == 0 && top == 0);
		init_size(r, true, false);
		decodeCodec(b, r, _codec37);
		_codec37_called = true;
		break;
	case 47:
		init_size(r, false, true);
		decodeCodec(b, r, _codecd);
		break;
	case 21:
	case 44:
		init_size(r, true, true);
		decodeCodec(b, r, _codec44);
		break;
	default:
		error("Invalid codec for frame object : %d", (int)codec);
	}
}


void smush_player::handleFrame(chunck & b) {
	checkBlock(b, TYPE_FRME);
	debug(6, "smush_player::handleFrame(%d)", _frame);
	_already_init = false;
	_skipnext = false;
	
	while(!b.eof()) {
		chunck * sub = b.sub_block();
		if(sub->get_size() & 1) b.seek(1);
		switch(sub->get_type()) {
			case TYPE_NPAL:
				handleNewPalette(*sub);
				break;
			case TYPE_FOBJ:
				handleFrameObject(*sub);
				break;
			case TYPE_PSAD:
				handleSoundFrame(*sub);
				break;
			case TYPE_TRES:
				handleTextResource(*sub);
				break;
			case TYPE_XPAL:
				handleDeltaPalette(*sub);
				break;
			case TYPE_IACT:
				handleImuseAction(*sub);
				break;
			case TYPE_STOR:
				handleStore(*sub);
				break;
			case TYPE_FTCH:
				handleFetch(*sub);
				break;
			case TYPE_SKIP:
				handleSkip(*sub);
				break;
			default:
				error("Unknown frame subchunck found : %s, %d", chunck::chunck_string(sub->get_type()), sub->get_size());
		}
		delete sub;
	}
	if(_cur_buffer) { _renderer->unlock_frame(); _cur_buffer = 0; }
	if(_output_sound) _mixer->handle_frame();
#ifdef DEBUG
	debug(5, "===================END OF FRAME========================"); 
#endif
	_renderer->flip_frame();
	if(_wait) _renderer->wait(WAIT);
	_frame++;
}

void smush_player::handleAnimHeader(chunck & b) {
	checkBlock(b, TYPE_AHDR, 774); 
	debug(6, "smush_player::handleAnimHeader()");
	_version = b.get_word();
	_nbframes = b.get_word();
	int unknown = b.get_word();
#ifdef DEBUG
	debug(5, "SMUSH HEADER : version == %d, nbframes == %d, unknown == %d", _version, _nbframes, unknown);
#else
	unknown = unknown;
#endif
	_renderer->start_decode(_fname, _version, _nbframes);
	readPalette(_pal, b);
	update_palette();
	if(_version == 1) {
		_sound_frequency = 22050;
	}
	if(_version == 2) {
		_secondary_version = b.get_dword();
		int unknown2 = b.get_dword();
		_sound_frequency = b.get_dword();
#ifdef DEBUG
		debug(5, "SMUSH HEADER : secondary version == %d, unknown2 == %d, sound frequency == %d", _secondary_version, unknown2, _sound_frequency);
		int i = 0, c;
		while(!b.eof()) {
			c = b.get_byte();
			if(c) debug(9, "SMUSH HEADER : remaining bytes : %d == %d", i, c);
			i++;
		}
#else
		unknown2 = unknown2;
#endif
		if(_secondary_version != 10 && _secondary_version != 0 && _secondary_version != 12 && _secondary_version != 15 && _secondary_version != 14)
			error("Wrong secondary version number for SMUSH animation");
		if(_sound_frequency != 0 && _sound_frequency != 11025 && _sound_frequency != 22050)
			error("Wrong _sound_frequency number for SMUSH animation");
	}else if(_version > 2) {
		error("Wrong primary version number for SMUSH animation");
	}
	if(_output_sound && _sound_frequency) {
		if(_sound_frequency != 22050) _sound_frequency = 22050;
		_mixer = _renderer->get_mixer();
		if(_mixer) {
			_mixer->init();
		} else {
			_output_sound = false;
		}
	}
}

static string_resource * get_strings(const char * file, bool is_encoded) {
	debug(7, "trying to read text ressources from %s", file);
	FILE * is;
	is = fopen(file, "rb");
	if(is == NULL) return 0;
	fseek(is, 0, SEEK_END);
	int length = ftell(is);
	fseek(is, 0, SEEK_SET);
	char * filebuffer = new char [length+1];
	assert(filebuffer);
	fread (filebuffer,length,1, is);
	filebuffer[length] = 0;
	fclose(is);
	if(is_encoded) {
		static const int ETRS_HEADER_LENGTH = 16;
		assert(length > ETRS_HEADER_LENGTH);
		chunck::type type = READ_BE_UINT32(filebuffer);
		if(type != TYPE_ETRS) error("invalid type for file"); // mem leak !!!
		char * old = filebuffer;
		filebuffer = new char[length - ETRS_HEADER_LENGTH];
		for(int i = ETRS_HEADER_LENGTH; i < length; i++)
			filebuffer[i-ETRS_HEADER_LENGTH] = old[i] ^ 0xCC;
		delete []old;
		length -= ETRS_HEADER_LENGTH;
	}
	string_resource * sr = new string_resource;
	assert(sr);
	sr->init(filebuffer, length);
	delete []filebuffer;
	return sr;
}

bool smush_player::read_string(const char * file, bool & ft) {
	const char * i = strrchr(file, '.');
	if(i == NULL) error("invalid filename : %s", file);
	char fname[260];
	memcpy(fname, file, i - file);
	strcpy(fname+(i-file), ".trs");
	if((_strings = get_strings(fname, false)) != 0) {
		ft = true;
		return true;
	}
	i = strrchr(file, '\\');
	if(i == NULL) i = strrchr(file, '/');
	else {
		char * j = strrchr(file, '/');
		if(j > i) i = j;
	}
	if(i == NULL) error("invalid filename : %s", file);
/*
	// unneeded, mineroad should be handled directly in scumm.
	memcpy(fname, file, i - file +1);
	strcpy(fname+(i-file+1), "mineroad.trs");
	if((_strings = get_strings(fname, false)) != 0) {
		ft = true;
		return true;
	}
*/
	memcpy(fname, file, i - file +1);
	strcpy(fname+(i-file+1), "digtxt.trs");
	if((_strings = get_strings(fname, true)) != 0) {
		ft = false;
		return true;
	}
	return false;
}

static font_renderer * load_font(const char * file, bool original = false) {
#ifdef DEBUG
	debug(5, "loading font from \"%s\"", file);
#endif
	font_renderer * fr = new font_renderer(original);
	smush_player p(fr, false, false);
	p.play(file);
	return fr;
}

bool smush_player::play(const char * file) {
#ifdef DEBUG
	debug(5, "start of animation : %s", file);
#endif	
	char * i = strrchr(file, '\\');
	if(i == NULL)
	{
		i = strrchr(file, '/');
	} else {
		char * j = strrchr(i, '/');
		if(j != NULL)
			i = j;
	}
	char directory[260];
	if(i != NULL) {
		strcpy(directory, file);
		directory[i-file] = 0;
		//! @todo remove this...
		_fname = strdup(i);
	} else {
		directory[0] = 0;
		_fname = strdup(file);
	}
	clean();
	
	if(_wait) {
		bool is_fullthrottle;
		if(!read_string(file, is_fullthrottle))
			warning("unable to read text information for \"%s\"", file);
		if(_strings) {
			if(is_fullthrottle) {
				if(strcmp(directory, "") == 0) {
					strcpy(directory, "../data/");
				} else {
					char * i = strrchr(directory, '\\');
					char * j = strrchr(directory, '/');
					if(j > i) i = j;
					if(i == NULL) {
						strcpy(directory, "data/");
					} else {
						*i = 0;
						strcat(directory, "/data/");
					}
				}
				char file[260];
				strcpy(file, directory); strcat(file, "scummfnt.nut");
				_fr[0] = load_font(file, true);
				strcpy(file, directory); strcat(file, "titlfnt.nut");
				_fr[2] = load_font(file, true);
			} else {
				for(int i = 0; i < 4; i++) {
					char file[260];
					sprintf(file, "%s/font%d.nut",directory, i);
					_fr[i] = load_font(file, i != 0);
				}
			}
		}
	}
	file_chunck base = file_chunck(file);
	
	checkBlock(base, TYPE_ANIM); 
	
	while(!base.eof()) {
		chunck * sub = base.sub_block();
		switch(sub->get_type()) {
			case TYPE_AHDR:
				handleAnimHeader(*sub);
				break;
			case TYPE_FRME:
				handleFrame(*sub);
				break;
			default:
				error("Unknown chunck found : %d, %d", sub->get_type(), sub->get_size());
		}
		delete sub;
		if(_renderer->premature_close())
			break;
	}
#ifdef DEBUG
	debug(5, "end of animation");
#endif	
	if(_output_sound) {
		_mixer->stop();
	}
	return true;
}
