/* ScummVM - Scumm Interpreter
 * Copyright (C) 2001/2002 The ScummVM project
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Header$
 *
 */

#ifndef __RECT_H_
#define __RECT_H_

#include "config.h"

/*! 	@brief simple class for handling both 2D position and size

	This small class is an helper for position and size values.
*/
class point {
private:
	int _x;	//!< The horizontal part of the point
	int _y;	//!< The vertical part of the point
public:
	point() : _x(0), _y(0) {};
	point(const point & p) : _x(p.get_x()), _y(p.get_y()) {};
	explicit point(int x, int y) : _x(x), _y(y) {};
	point & operator=(const point & p) { _x = p.get_x(); _y = p.get_y(); return *this; };
	bool operator==(const point & p) const { return _x == p.get_x() && _y == p.get_y(); };
	const int & get_x() const { return _x; };
	const int & get_y() const { return _y; };
	int & get_x() { return _x; };
	int & get_y() { return _y; };
	point operator+(const point & p) const { return point(_x+p.get_x(), _y+p.get_y()); };
	point operator-(const point & p) const { return point(_x-p.get_x(), _y-p.get_y()); };
	point & operator+=(const point & p) { _x += p.get_x(); _y += p.get_y(); return *this; };
	point & operator-=(const point & p) { _x -= p.get_x(); _y -= p.get_y(); return *this; };
	bool is_origin() const { return *this == point(0, 0); };
	void set(int x, int y) { _x = x; _y = y; }
};

/*! 	@brief simple class for handling a rectangular zone.

	This small class is an helper for rectangles.
	It is mostly used by the blitter class.
*/
class rect {
private:
	point _topleft;		//!< The point at the top left of the rectangle
	point _bottomright;	//!< The point at the bottom right of the rectangle
protected:
	void check();
public:
	rect();
	rect(int x, int y);
	explicit rect(const point & size);
	rect(int x1, int y1, int x2, int y2);
	rect(const point & topleft, const point & bottomright);
	rect(const rect & r);
	rect & operator=(const rect & r);
	bool operator==(const rect & r) const;
	point size() const { return (_bottomright - _topleft); };
	int width() const { return size().get_x(); }
	int height() const { return size().get_y(); }
	int left() const { return _topleft.get_x(); }
	int right() const { return _bottomright.get_x(); }
	int top() const { return _topleft.get_y(); }
	int bottom() const { return _bottomright.get_y(); }
	const point & topleft() const { return _topleft; }
	const point & bottomright() const { return _bottomright; }
	
	/*!	@brief check if given position is inside the rectangle
		
		@param x the horizontal position to check
		@param y the vertical position to check	
		
		@return true if the given position is inside the rectangle, false otherwise
	*/
	bool is_inside(int x, int y) const;
	/*!	@brief check if given point is inside the rectangle
		
		@param p the point to check
		
		@return true if the given point is inside the rectangle, false otherwise
	*/
	bool is_inside(const point & p) const;
	bool clip(rect & r) const;
};

#endif
