/* ScummVM - Scumm Interpreter
 * Copyright (C) 2001/2002 The ScummVM project
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Header$
 *
 */

#include <stdafx.h>
#include "channel.h"
#include "chunck.h"
#include "chunck_type.h"

#include <assert.h>
#include <string.h> // for memcpy.h
#ifndef min
#define min(x, y) ((x) > (y) ? (y) : (x))
#endif

void saud_channel::handleStrk(chunck & b) {
	int size = b.get_size();
	if(size != 14 && size != 10) {
		error("STRK has a invalid size : %d", size);
	}
}

void saud_channel::handleSmrk(chunck & b) {
	_mark_reached = true;
}

void saud_channel::handleShdr(chunck & b) {
	int size = b.get_size();
	if(size != 4) warning("SMRK has a invalid size : %d", size);
}

bool saud_channel::handleSubTags(int & offset) {
	int available_size = _tbuffer_size - offset;
	if(available_size >= 8) {
		chunck::type type = READ_BE_UINT32(_tbuffer + offset);
		unsigned int size = READ_BE_UINT32(_tbuffer + offset + 4);
		switch(type) {
			case TYPE_STRK:
				_in_data = false;
				if(available_size >= (size + 8)) {
					cont_chunck c((char*)_tbuffer + offset);
					handleStrk(c);
				} else return false;
				break;
			case TYPE_SMRK:
				_in_data = false;
				if(available_size >= (size + 8)) {
					cont_chunck c((char*)_tbuffer + offset);
					handleSmrk(c);
				} else return false;
				break;
			case TYPE_SHDR:
				_in_data = false;
				if(available_size >= (size + 8)) {
					cont_chunck c((char*)_tbuffer + offset);
					handleShdr(c);
				} else return false;
				break;
			case TYPE_SDAT: // Sound data !!!
				_in_data = true;
				_data_size = size;
				offset += 8;
				return false;
			default:
				error("unknown chunck in SAUD track : %s ", chunck::chunck_string(type));
		}
		offset += size + 8;
		return true;
	}
	return false;
}

bool saud_channel::processBuffer() {
	// At the start of this function, we have _tbuffer[0.._tbuffersize] containing possible data...
	// and _sbuffer is 0
	// At the end we have :
	// if(sound data) _sbuffer[0.._sbuffer_size] contains the sound data
	// the unprocessed data is kept in _tbuffer[0.._tbuffersize] (which may have changed)
	// if no unprocessed data, then _tbuffer is 0
	assert(_tbuffer != 0);
	assert(_tbuffer_size != 0);
	assert(_sbuffer == 0);
	assert(_sbuffer_size == 0);
	
	if(_in_data) {
		if(_data_size < _tbuffer_size) {
			// I can't assume that the channel is finished after data is received... (this assumption failed in realride.san)
			int offset= _data_size;
			while(handleSubTags(offset));
			_sbuffer_size = _data_size;
			_sbuffer = _tbuffer;
			if(offset < _tbuffer_size) { // there is still some unprocessed data
				int new_size = _tbuffer_size - offset;
				_tbuffer = new unsigned char[new_size];
				if(!_tbuffer)  error("saud_channel failed to allocate memory");
				memcpy(_tbuffer, _sbuffer + offset, new_size);
				_tbuffer_size = new_size;
			} else {
				_tbuffer = 0;
				_tbuffer_size = 0;
			}
			if(_sbuffer_size == 0) {
				// this never happened yet, but who knows
				delete []_sbuffer; 
				_sbuffer = 0;
			}
		} else {
			// easy, swap the buffer
			_sbuffer_size = _tbuffer_size;
			_sbuffer = _tbuffer;
			_tbuffer_size = 0;
			_tbuffer = 0;
		}
	} else {
		int offset = 0;
		while(handleSubTags(offset));
		//~ assert(_in_data);
		if(_in_data) {
			//~ unsigned char * old = _tbuffer;
			_sbuffer_size = _tbuffer_size - offset;
			assert(_sbuffer_size);
			_sbuffer = new unsigned char[_sbuffer_size];
			if(!_sbuffer)  error("saud_channel failed to allocate memory");
			memcpy(_sbuffer, _tbuffer + offset, _sbuffer_size);
			delete []_tbuffer;
			_tbuffer = 0;
			_tbuffer_size = 0;
		} else {
			if(offset) { // maybe I should assert() this to avoid a lock...
				unsigned char * old = _tbuffer;
				int new_size = _tbuffer_size - offset;
				_tbuffer = new unsigned char[new_size];
				if(!_tbuffer)  error("saud_channel failed to allocate memory");
				memcpy(_tbuffer, old + offset, new_size);
				_tbuffer_size = new_size;
				delete []old;
			}
		}
	}
	return true;
}

saud_channel::saud_channel(int track, int freq) : 
			_track(track), 
			_nbframes(0),
			_data_size(-1),
			_tbuffer(0), 
			_sbuffer(0), 
			_frequency(freq),
			_tbuffer_size(0), 
			_sbuffer_size(0), 
			_in_data(false),
			_mark_reached(false) {
}

saud_channel::~saud_channel() {
	if(_tbuffer) delete []_tbuffer;
	if(_sbuffer) {
		warning("this should never happen !!!! (_sbuffer not NULL here)");
		delete []_sbuffer;
	}
}

bool saud_channel::is_terminated() const {
	return (_mark_reached && _data_size == 0 && _sbuffer == 0);
}

void saud_channel::recalcVolumeTable() {
	const int MAX_BALANCE = 100;
	int volume_left, volume_right;
	if(_balance < -MAX_BALANCE || _balance > MAX_BALANCE) {
		error("balance is out of range ! : %d", _balance);
	}
	int left_multiplier = MAX_BALANCE - _balance;
	int right_multiplier = MAX_BALANCE + _balance;
	volume_left = _volume * left_multiplier / (MAX_BALANCE*2);
	volume_right = _volume * right_multiplier / (MAX_BALANCE*2);
	if(volume_left < 0) volume_left = 0;
	if(volume_left > 128) volume_left = 128;
	if(volume_right < 0) volume_right = 0;
	if(volume_right > 128) volume_right = 128;
	for(int i = 0; i < 256; i++) {
		int value = volume_left * (signed char)i;
		_voltable[0][i] = TO_BE_16(value);
		value = volume_right * (signed char)i;
		_voltable[1][i] = TO_BE_16(value);
	}
}

bool saud_channel::set_parameters(int nb, int flags, int volume, int balance) {
	_nbframes = nb;
	_flags = flags; // bit 7 == IS_VOICE, bit 6 == IS_BACKGROUND_MUSIC, other ??
	_volume = volume;
	_balance = balance;
	_index = 0;
	recalcVolumeTable();
	return true;
}

bool saud_channel::check_parameters(int index, int nb, int flags, int volume, int balance) {
	if(++_index != index) error("invalid index in saud_channel::check_parameters()");
	if(_nbframes != nb) error("invalid duration in saud_channel::check_parameters()");
	if(_flags != flags) error("invalid flags in saud_channel::check_parameters()");
	if(_volume != volume || _balance != balance) {
		_volume = volume;
		_balance = balance;
		recalcVolumeTable();
	}
	return true;
}

bool saud_channel::append_data(chunck & b, int size) {
	if(_data_size == -1) { // First call
		assert(size > 8);
		chunck::type saud_type = b.get_dword(); saud_type = TO_BE_32(saud_type);
		unsigned int saud_size = b.get_dword(); saud_size = TO_BE_32(saud_size);
		if(saud_type != TYPE_SAUD) error("Invalid CHUNCK for saud_channel : %X", saud_type);
		size -= 8;
		_data_size = -2; // We don't get here again...
	}
	if(_tbuffer) {
		unsigned char * old = _tbuffer;
		_tbuffer = new unsigned char[_tbuffer_size + size];
		if(!_tbuffer)  error("saud_channel failed to allocate memory");
		memcpy(_tbuffer, old, _tbuffer_size);
		delete []old;
		b.read(_tbuffer + _tbuffer_size, size);
		_tbuffer_size += size;
	} else {
		_tbuffer_size = size;
		_tbuffer = new unsigned char[_tbuffer_size];
		if(!_tbuffer)  error("saud_channel failed to allocate memory");
		b.read(_tbuffer, _tbuffer_size);
	}
	return processBuffer();
}

int saud_channel::available_sound_data(void) const {
	return _sbuffer_size;
}

void saud_channel::get_sound_data(short * snd, int size) {
	for(int i = 0; i < size; i++) {
		snd[2*i] = _voltable[0][_sbuffer[i] ^ 0x80];
		snd[2*i+1] = _voltable[1][_sbuffer[i] ^ 0x80];
	}
	_data_size -= size;
	delete []_sbuffer;
	_sbuffer = 0;
	_sbuffer_size = 0;
}
