#include <stdafx.h>
#include "scumm_renderer.h"
#include "channel.h"

class scumm_mixer : public mixer {
private:
	static const int MAX_CHANNELS = SoundMixer::NUM_CHANNELS;
	SoundMixer * _mixer; //!< pointer to the SoundMixer instance
	struct {
		int id;
		channel * chan;
		bool first;
		int mixer_index;
	} _channels[MAX_CHANNELS];		//!< The map of track and channels
	int _next_index;
public:
	scumm_mixer(SoundMixer *);
	virtual ~scumm_mixer();
	bool init();
	channel * find_channel(int track);
	bool add_channel(channel * c);
	bool handle_frame();
	bool stop();
	bool update();
};

scumm_mixer::scumm_mixer(SoundMixer * m) : _mixer(m), _next_index(0) {
	for(int i = 0; i < MAX_CHANNELS; i++) {
		_channels[i].id = -1;
		_channels[i].chan = 0;
		_channels[i].first = true;
	}
}

scumm_mixer::~scumm_mixer() {
}

bool scumm_mixer::init() {
	debug(9, "scumm_mixer::init()");
	return true;
}

channel * scumm_mixer::find_channel(int track) {
	debug(9, "scumm_mixer::find_channel(%d)", track);
	for(int i = 0; i < MAX_CHANNELS; i++) {
		if(_channels[i].id == track)
			return _channels[i].chan;
	}
	return 0;
}

bool scumm_mixer::add_channel(channel * c) {
	int track = c->get_track_identifier();
	debug(9, "scumm_mixer::add_channel(%d)", track);
//~ #ifdef DEBUG_MIXER
	for(int i = 0; i < MAX_CHANNELS; i++) {
		if(_channels[i].id == track)
			warning("mixer::add_channel(%d) : channel already exist !", track);
	}
//~ #endif
	if(_next_index >= MAX_CHANNELS) _next_index = 0;
	for(int i = _next_index; i < MAX_CHANNELS; i++) {
		if(_channels[i].chan == 0 || _channels[i].id == -1)
		{
			_channels[i].chan = c;
			_channels[i].id = track;
			_channels[i].first = true;
			_next_index = i +1;
			return true;
		}
	}
	for(int i = 0; i < _next_index; i++) {
		if(_channels[i].chan == 0 || _channels[i].id == -1)
		{
			_channels[i].chan = c;
			_channels[i].id = track;
			_channels[i].first = true;
			_next_index = i +1;
			return true;
		}
	}
	fprintf(stderr, "_next_index == %d\n", _next_index);
	for(int i = 0; i < MAX_CHANNELS; i++) {
		fprintf(stderr, "channel %d : %p(%d, %d) %d %d\n", i, _channels[i].chan, 
			_channels[i].chan ? _channels[i].chan->get_track_identifier() : -1, 
			_channels[i].chan ? _channels[i].chan->is_terminated() : 1, 
			_channels[i].first, _channels[i].mixer_index);
	}
	error("mixer::add_channel() : no more channel available");
	return false;
}

bool scumm_mixer::handle_frame() {
	debug(9, "scumm_mixer::handle_frame()");
	for(int i = 0; i < MAX_CHANNELS; i++) {
		if(_channels[i].id != -1)
		{
			debug(9, "updating channel %d (%p)", _channels[i].id, _channels[i].chan);
			if(_channels[i].chan->is_terminated()) {
				debug(9, "channel %d has terminated (%p)", _channels[i].id, _channels[i].chan);
				delete _channels[i].chan;
				_channels[i].id = -1;
				_channels[i].chan = 0;
			} else {
				int rate;
				bool stereo, is_short;
				_channels[i].chan->get_parameters(rate, stereo, is_short);
				int size = _channels[i].chan->available_sound_data();
				debug(9, "channel %d : %d, %s, %d bits, %d", _channels[i].id, rate, stereo ? "stereo" : "mono", is_short ? 16 : 8, size);
				int flags = stereo ? SoundMixer::FLAG_STEREO : 0;
				if(is_short) {
					// FIXME this is one more data copy... we could get rid of it...
					short * data = new short[size*(stereo ? 2 : 1)];
					_channels[i].chan->get_sound_data(data, size);
					size *= stereo ? 4 : 2;
					// append to _sound
					if(_channels[i].first) {
						_channels[i].mixer_index = _mixer->play_stream(NULL, -1, data, size, rate, flags | SoundMixer::FLAG_16BITS);
						debug(5, "channel %d bound to mixer_index %d", _channels[i].id, _channels[i].mixer_index);
						_channels[i].first = false;
					} else {
						_mixer->append(_channels[i].mixer_index, data, size, rate, flags | SoundMixer::FLAG_16BITS);
					}
					delete []data;
				} else {
					char * data = new char[size*(stereo ? 2 : 1)];
					_channels[i].chan->get_sound_data(data, size);
					size *= stereo ? 2 : 1;
					// append to _sound
					if(_channels[i].first) {
						_channels[i].mixer_index = _mixer->play_stream(NULL, -1, data, size, rate, flags | SoundMixer::FLAG_UNSIGNED);
						_channels[i].first = false;
					} else {
						_mixer->append(_channels[i].mixer_index, data, size, rate, flags | SoundMixer::FLAG_UNSIGNED);
					}
					delete []data;
				}
			}
		}
	}
	return true;
}

bool scumm_mixer::stop() {
	debug(9, "scumm_mixer::stop()");
	for(int i = 0; i < MAX_CHANNELS; i++) {
		if(_channels[i].id != -1) {
				delete _channels[i].chan;
				_channels[i].id = -1;
				_channels[i].chan = 0;
		}
	}
	//~ _mixer->stop_all();
	return true;
}

scumm_renderer::scumm_renderer(Scumm * scumm) : _scumm(scumm), _smixer(0) {
}

static scumm_renderer * s_renderer;

static void smush_handler(Scumm * scumm) {
	s_renderer->update();
}

mixer * scumm_renderer::get_mixer() {
	if(_smixer == 0) {
		_scumm->_sound->pauseBundleMusic(true);
		_smixer = new scumm_mixer(_scumm->_mixer);
		if(!_smixer) error("unable to allocate a smush mixer");
		s_renderer = this;
		_scumm->_timer->installProcedure(&smush_handler, 75);
	}
	return _smixer;
}

scumm_renderer::~scumm_renderer() {
	_scumm->_insaneState = 0;
	_scumm->exitCutscene();
	if(_smixer) {
		_scumm->_timer->releaseProcedure(&smush_handler);
		delete _smixer;
		_smixer = 0;
	}
	_scumm->_sound->pauseBundleMusic(false);
}

bool scumm_renderer::wait(int ms) {
	while(_wait) {
		_scumm->waitForTimer(1);
	}
	return true; 
}

bool scumm_renderer::start_decode(const char * fname, int version, int nbframes) {
	_scumm->_sound->pauseBundleMusic(true);
	_scumm->videoFinished = 0;
	_scumm->_insaneState = 1;
	return true;
}

bool scumm_renderer::set_palette(const palette & pal) {
	int32 i;
	byte palette_colors[1024];
	byte *p = palette_colors;

	for (i = 0; i < 256; i++, p += 4) {
		p[0] = pal[i].red();
		p[1] = pal[i].green();
		p[2] = pal[i].blue();
		p[3] = 0;
	}

	_scumm->_system->set_palette(palette_colors, 0, 256);
	_scumm->setDirtyColors(0, 255);
	return base_renderer::set_palette(pal); // For compatibility with possible subclass...
}

void scumm_renderer::save(int frame) {
	int width = min(get_width(), _scumm->_realWidth); 
	int height = min(get_height(), _scumm->_realHeight);
	
	_scumm->_system->copy_rect((const byte*)data(), get_width(), 0, 0, width, height);
	_scumm->_system->update_screen();
	_scumm->processKbd();
	_wait = true;
}

bool scumm_renderer::premature_close() { 
	return _scumm->videoFinished; 
}

bool scumm_renderer::update() {
	_wait = false;
}


